import cffi
ffibuilder = cffi.FFI()

header = """
extern void fem(void);
"""

module = """
from my_plugin import ffi
from dolfin import *

@ffi.def_extern()
def fem(): 
    # Sub domain for no-slip (mark whole boundary, inflow and outflow will overwrite)
    class Noslip(SubDomain):
        def inside(self, x, on_boundary):
            return on_boundary

    # Sub domain for inflow (right)
    class Inflow(SubDomain):
        def inside(self, x, on_boundary):
            return x[0] > 1.0 - DOLFIN_EPS and on_boundary

    # Sub domain for outflow (left)
    class Outflow(SubDomain):
        def inside(self, x, on_boundary):
            return x[0] < DOLFIN_EPS and on_boundary
    
    # Read mesh
    mesh = Mesh("dolfin_fine.xml.gz")
    
    # Create mesh functions over the cell facets
    sub_domains = MeshFunction("size_t", mesh, mesh.topology().dim() - 1)
    sub_domains_bool = MeshFunction("bool", mesh, mesh.topology().dim() - 1)
    sub_domains_double = MeshFunction("double", mesh, mesh.topology().dim() - 1)
    
    # Mark all facets as sub domain 3
    sub_domains.set_all(3)
    sub_domains_bool.set_all(False)
    sub_domains_double.set_all(0.3)
    
    # Mark no-slip facets as sub domain 0, 0.0
    noslip = Noslip()
    print("Here")
    noslip.mark(sub_domains, 0)
    print("There")
    noslip.mark(sub_domains_double, 0.0)
    
    # Mark inflow as sub domain 1, 01
    inflow = Inflow()
    inflow.mark(sub_domains, 1)
    inflow.mark(sub_domains_double, 0.1)
    
    # Mark outflow as sub domain 2, 0.2, True
    outflow = Outflow()
    outflow.mark(sub_domains, 2)
    outflow.mark(sub_domains_double, 0.2)
    outflow.mark(sub_domains_bool, True)
    
    # Save sub domains to file
    file = File("subdomains.xml")
    file << sub_domains
    
    # FIXME: Not implemented
    #file_bool = File("subdomains_bool.xml")
    #file_bool << sub_domains_bool
    
    file_double = File("subdomains_double.xml")
    file_double << sub_domains_double
    
    # Save sub domains to VTK files
    file = File("subdomains.pvd")
    file << sub_domains
    
    file = File("subdomains_double.pvd")
    file << sub_domains_double
         


"""

with open("plugin.h", "w") as f:
    f.write(header)

ffibuilder.embedding_api(header)
ffibuilder.set_source("my_plugin", r'''
    #include "plugin.h"
''')

ffibuilder.embedding_init_code(module)
ffibuilder.compile(target="libplugin.so", verbose=True)
